/*------------------------------------------------*/
/* 07osvetleni/normals.c                          */

#include "normals.h"
#include "linmath.h"
#include <stdlib.h>
#include <stdio.h>

static void cn_free(const ComputedNormals * cn)
{
    free(cn->normals);
}

ComputedNormals createNormals(const GLfloat * vertices,
			      size_t verticesLength, const int *directions)
{
    ComputedNormals cn;
    cn.length = verticesLength - (verticesLength % 9);
    cn.free = cn_free;
    cn.normals = (GLfloat *) malloc(sizeof(GLfloat) * cn.length);
    size_t i;
    vec3 n, v1, v2;
    for (i = 0; i <= verticesLength - 9; i += 9) {
	size_t j = i / 9;
	const GLfloat *a, *b, *c;
	a = &vertices[i];
	b = &vertices[i + 3];
	c = &vertices[i + 6];
	vec3_sub(v1, a, b);
	vec3_sub(v2, b, c);
	if (directions == NULL || directions[j] != 0) {
	    vec3_mul_cross(n, v1, v2);
	} else {
	    vec3_mul_cross(n, v2, v1);
	}
	vec3_norm(n, n);
	for (j = 0; j < 3; j++) {
	    cn.normals[i + j * 3 + 0] = n[0];
	    cn.normals[i + j * 3 + 1] = n[1];
	    cn.normals[i + j * 3 + 2] = n[2];
	}
    }
    return cn;
}

ComputedNormals createNormalsByIndex(const GLfloat * vertices,
				     size_t verticesLength,
				     const GLuint * indices,
				     size_t indicesLength,
				     const int *directions)
{
    ComputedNormals cn;
    cn.length = indicesLength * 3;
    cn.free = cn_free;
    cn.normals = (GLfloat *) malloc(sizeof(GLfloat) * cn.length);

    size_t ii;
    vec3 n, v1, v2;
    for (ii = 0; ii <= indicesLength-3; ii += 3) {
	size_t j = ii / 3;
	size_t i1 = indices[ii]* 3;
	size_t i2 = indices[ii + 1] * 3;
	size_t i3 = indices[ii + 2] * 3;
	vec3 a;
	a[0] = vertices[i1];
	a[1] = vertices[i1 + 1];
	a[2] = vertices[i1 + 2];
	vec3 b;
	b[0] = vertices[i2];
	b[1] = vertices[i2 + 1];
	b[3] = vertices[i2 + 2];
	vec3 c;
	c[0] = vertices[i3];
	c[1] = vertices[i3 + 1];
	c[2] = vertices[i3 + 2];
	vec3_sub(v1, a, b);
	vec3_sub(v2, b, c);
	if (directions == NULL || directions[j] != 0) {
	    vec3_mul_cross(n, v1, v2);
	} else {
	    vec3_mul_cross(n, v2, v1);
	}
	vec3_norm(n, n);

	cn.normals[i1] = n[0];
	cn.normals[i1 + 1] = n[1];
	cn.normals[i1 + 2] = n[2];
	cn.normals[i2] = n[0];
	cn.normals[i2 + 1] = n[1];
	cn.normals[i2 + 2] = n[2];
	cn.normals[i3] = n[0];
	cn.normals[i3 + 1] = n[1];
	cn.normals[i3 + 2] = n[2];
    }
    return cn;
}

/*------------------------------------------------*/
